/*************************************************************************
 *  playmus.c
 *
 * New code (C) 2006 Andrew J. Kroll <a@oo.ms>
 *
 * Original Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include "musserver.h"
#include <errno.h>
/* fixup for stupid glibc */
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <sys/msg.h>


void seqbuf_dump(void);
void reset_midi(void);
void note_off(int note, int channel, int volume);
void note_on(int note, int channel, int volume);
void pitch_bend(int channel, int value);
void control_change(int controller, int channel, int value);
void patch_change(int patch, int channel);
void midi_wait(float time);
void midi_timer(int action);
void vol_change(int volume);
void Mcleanup(int status);
extern int qid;
extern int verbose;
extern int doomver;
extern int changevol;

#define START 0
#define STOP 1
#define CONT 2

int returnval = 0;

/* from readwad.c */
extern char *doom2names[];
extern char *doom1names[];

void quitmus()
{
	returnval = TERMINATED;
}

void do_nothing()
{
	signal(SIGHUP, do_nothing);
}

int get_mesg(int flags)
{
	int result;
	struct msgbuf *recv;
	int msize = 9;
	int done = 0;
	int x = 0;

	while (!done) {
		done = 1;
		recv = malloc(sizeof(long) + 9);
		recv->mtext[0] = 0;
		result = msgrcv(qid, recv, msize, 0, flags + MSG_NOERROR);
		if (result > 0) {
			if (verbose >= 2)
				printf("musserver: >ipc< errno = %d, result = %d, mtext = %s, qid = %d\n", errno, result, recv->mtext, qid);
			switch (recv->mtext[0]) {
			
				case 'q':
				case 'Q':
					free(recv);
					returnval = TERMINATED;
					return(TERMINATED);
					break;

				case 'v':
				case 'V':
					if (changevol)
						vol_change(recv->mtext[1]);
					if (verbose)
						printf("musserver: Volume change = %d\n", recv->mtext[1]);
					done = 0;
					break;

				case 'D':
				case 'd':
					for (x = 0; x < strlen(recv->mtext); x++)
						recv->mtext[x] = toupper(recv->mtext[x]);
					if (verbose >= 2) printf("musserver: Caller asking for resource '%s'\n",recv->mtext);

					switch(doomver) {

						case 0:
						case 1:
							for (x = 0; x <= 31; x++)
								if (!strncmp(recv->mtext, doom1names[x], strlen(doom1names[x])))
								break;
							if (x == 32)
								x = 50;
							break;

						case 2:
							for (x = 0; x <= 34; x++)
								if (!strncmp(recv->mtext, doom2names[x], strlen(doom2names[x])))
									break;
							if (x == 35)
								x = 50;
							break;

						default:
							x = 50;
							break;
					}
					if (x != 50) {
						returnval = 500 + x;
						return(returnval);
					} else {
						done = 0;
					}
					break;
			}
		} else if (result == -1) switch (errno) {

			case EACCES:
				printf("musserver: Could not receive IPC message: no read access permission!\n");
				Mcleanup(1);
				break;

			case EFAULT:
				printf("musserver: Could not receive IPC message: memory address is inaccessible.\n");
				Mcleanup(1);
				break;

			case EIDRM:
				Mcleanup(1);
				break;

			case EINTR:
				printf("musserver: Could not receive IPC message: received an interrupt signal.\n");
				break;

			case EINVAL:
				printf("musserver: Could not receive IPC message: invalid message size or queue id.\n");
				Mcleanup(1);
				break;

			case ENOMSG:
				break;
		}
		free(recv);
	}
	return(returnval);
}


int playmus(char *musdata, unsigned int mussize, int play_once)
{
	unsigned char event;
	unsigned char event1;
	unsigned char event2;
	unsigned char eventtype;
	unsigned char channelnum;
	unsigned char last = 0;
	unsigned char notenum;
	unsigned char notevol;
	unsigned int pitchwheel;
	unsigned char controlnum;
	unsigned char controlval;
	unsigned int ticks;
	unsigned char tmpch;
	unsigned int done = 0;
	unsigned int lastvol[16];
	double delaytime;
	double curtime = 0.0;
	char *muscopy;


	signal(SIGHUP, do_nothing);
	signal(SIGQUIT, quitmus);
	signal(SIGINT, quitmus);
	signal(SIGTERM, quitmus);
	signal(SIGCONT, SIG_IGN);

	returnval = 0;
	muscopy = musdata;

	if (play_once == 2) get_mesg(MSG_WAIT);


	reset_midi();
	midi_timer(START);


	while(!done) {
		if (!returnval) {
			returnval = get_mesg(IPC_NOWAIT);
		}
		if (returnval != 0)
			return(returnval);
		event = *musdata++;
		channelnum = event & 15;
		if (channelnum > 8)
			channelnum++;
		if (channelnum == 16)
			channelnum = 9;
		eventtype = (event >> 4) & 7;
		last = event & 128;

		switch (eventtype) {
			case 0:		/* note off */
				event1 = *musdata++;
				notenum = event1 & 127;
				note_off(notenum, channelnum, lastvol[channelnum]);
				break;

			case 1:		/* note on */
				event1 = *musdata++;
				notenum = event1 & 127;
				if (event1 & 128) {
					event2 = *musdata++;
					notevol = event2 & 127;
					lastvol[channelnum] = notevol;
				}
				note_on(notenum, channelnum, lastvol[channelnum]);
				break;

			case 2:		/* pitch wheel */
				event1 = *musdata++;
				pitchwheel = event1 / 2;
				pitch_bend(channelnum, pitchwheel);
				break;

			case 4:		/* midi controller change */
				event1 = *musdata++;
				controlnum = event1 & 127;
				event2 = *musdata++;
				controlval = event2 & 127;
				switch (controlnum) {

					case 0:		/* patch change */
						patch_change(controlval, channelnum);
						break;

					case 3:		/* volume */
						control_change(CTL_MAIN_VOLUME, channelnum, controlval);
						break;

					case 4:		/* pan */
						control_change(CTL_PAN, channelnum, controlval);
						break;
				}
				break;

			case 6:	/* end of music data */
				if (play_once)
					get_mesg(MSG_WAIT);
				else {
					usleep(100000); 
					musdata = muscopy;
					midi_timer(START);
					curtime = 0.0;
					last = 0;
				}
				break;

			case 3:	/* unknown, but contains data */
				musdata++;
				break;

			default:	/* unknown */
				break;
		}
		if (last) {	/* next data portion is time data */
			tmpch = *musdata++;
			ticks = tmpch & 127;
			while(tmpch & 128) {
				tmpch = *musdata++;
				ticks = (ticks * 128) + (tmpch & 127);
			}
			delaytime = (91.451955138 * (double)ticks) / 128.0;
			curtime += delaytime;
			midi_wait(curtime);
		}
	}
	return returnval;
}
